/*
**   IRC - Internet Relay Chat, source/commands.c
**   Copyright (C) 1999 Chris Behrens
**
**   Portions Copyright (C) 1990 Jarkko Oikarinen and
**                      University of Oulu, Computing Center
**
**   Portions Copyright (C) 2001 Adam Loghry
** 
**   This program is free software; you can redistribute it and/or modify
**   it under the terms of the GNU General Public License as published by
**   the Free Software Foundation; either version 1, or (at your option)
**   any later version.
**
**   This program is distributed in the hope that it will be useful,
**   but WITHOUT ANY WARRANTY; without even the implied warranty of
**   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**   GNU General Public License for more details.
**
**   You should have received a copy of the GNU General Public License
**   along with this program; if not, write to the Free Software
**   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

//#include "ircd_includes.h"
//#include "commands.h"
#include "cmds.h"
//#include "../lib/confparse/configparse.h"

RCSID("$Id: commands.c,v 1.20 2001/10/02 01:40:35 cbehrens Exp $");

extern int _user_modes;
extern int _channel_modes;

static void _build_channel_modes(IRCD_CHANNEL *chptr, IRCD_CLIENT *cptr, char *mbuf, char *pbuf);
static void _send_channel_modes(IRCD_CLIENT *cptr, IRCD_CHANNEL *chptr);
static void _umode_build_change(IRCD_CLIENT *pptr, char *buf, int oldmodes, int sendmask);
static void _umode_build(IRCD_CLIENT *pptr, char *buf, int mask);

static void _build_channel_modes(IRCD_CHANNEL *chptr, IRCD_CLIENT *cptr, char *mbuf, char *pbuf)
{
	int		ismem = 0;

	if (!IsServer(cptr) && (chptr->limit || chptr->key))
		ismem = (IsMember(chptr, cptr) != NULL);
	*mbuf++ = '+';
	if (chptr->modes & CHANNEL_MODES_SECRET)
		*mbuf++ = 's';
	/* else */
	if (chptr->modes & CHANNEL_MODES_PRIVATE)
		*mbuf++ = 'p';
	if (chptr->modes & CHANNEL_MODES_MODERATED)
		*mbuf++ = 'm';
	if (chptr->modes & CHANNEL_MODES_TOPICLIMIT)
		*mbuf++ = 't';
	if (chptr->modes & CHANNEL_MODES_INVITEONLY)
		*mbuf++ = 'i';
	if (chptr->modes & CHANNEL_MODES_NOPRIVMSGS)
		*mbuf++ = 'n';
	if (chptr->limit)
	{
		*mbuf++ = 'l';
		if (IsServer(cptr) || ismem)
			pbuf += sprintf(pbuf, "%d ", chptr->limit);
	}
	if (chptr->key)
	{
		*mbuf++ = 'k';
		if (IsServer(cptr) || ismem)
			pbuf += strlen(strcpy(pbuf, chptr->key));
	}
	*mbuf = '\0';
	*pbuf = '\0';
}

#if 0
static void _send_channel_modes(IRCD_CLIENT *cptr, IRCD_CHANNEL *chptr)
{
	char				modebuf[64];
	char				parabuf[512];
	char				buf[512];
	char				*t;
	IRCD_CLIENTLIST		*clptr;
	IRCD_CLIENTLIST		*anop = NULL;
	IRCD_CLIENTLIST		*skip = NULL;
	IRCD_BAN			*bans;
	IRCD_EXCEPT			*excepts;
	IRCD_QUIET			*quiets;
	int					n = 0;
	int					paralen;
	int					namlen;
	char				*cp;
	int					count;
	int					send;

	if (*chptr->name != '#') /* don't send '&' channels */
		return;
	_build_channel_modes(chptr, cptr, modebuf, parabuf);

	t = buf + sprintf(buf, ":%s SJOIN %u %s %s %s%s:",
			meptr->name, chptr->creation, chptr->name, modebuf,
			parabuf, *parabuf ? " " : "");

	for(clptr=chptr->members;clptr;clptr=clptr->next)
	{
		if (clptr->modes & CHANNEL_MODES_CHANOP)
		{
			anop = clptr;
			break;
		}
	}
	clptr = NULL;
	for(;;)
	{
		if (anop)
		{
			clptr = skip = anop;
			anop = NULL;
		}
		else
		{
			if (clptr == NULL || clptr == skip)
				clptr = chptr->members;
			else
				clptr = clptr->next;
			if (clptr && clptr == skip)
				clptr = clptr->next;
			if (clptr == NULL)
				break;
		}
		if (clptr->modes & CHANNEL_MODES_CHANOP)
			*t++ = '@';
		if (clptr->modes & CHANNEL_MODES_VOICE)
			*t++ = '+';
		strcpy(t, clptr->cptr->name);
		t += strlen(t);
		*t++ = ' ';
		n++;
		if (t - buf > 512 - HOSTLEN - 3)
		{
			*t++ = '\0';
			if (t[-1] == ' ')
				t[-1] = '\0';
			send_to_client(cptr, "%s", buf);
			t = buf + sprintf(buf, ":%s SJOIN %u %s 0 :",
					meptr->name, chptr->creation, chptr->name);
			n = 0;
		}
	}
	if (n)
	{
		*t++ = '\0';
		if (t[-1] == ' ')
			t[-1] = '\0';
		send_to_client(cptr, "%s", buf);
	}
	*parabuf = '\0';
	*modebuf = '+';
	cp = modebuf + 1;
	count = 0;
	paralen = 0;
	send = 0;
		
	/* send bans */

	for(bans=chptr->bans;bans;bans=bans->next)
	{
		if (paralen + (namlen = strlen(bans->banstring)) + 10 < 200)
		{
			*(parabuf+paralen) = ' ';
			strcpy(parabuf+paralen+1, bans->banstring);
			paralen += namlen + 1;
			count++;
			*cp++ = 'b';
		}
		else if (*parabuf)
			send = 1;
		if (count == 4)
			send = 1;
		if (send)
		{
			*cp = '\0';
			send_prefix_to_client(meptr, cptr, "MODE %s %s %s",
					chptr->name, modebuf, parabuf);
			send = 0;
			*parabuf = '\0';
			paralen = 0;
			cp = modebuf;
			*cp++ = '+';
			if (count != 4)
			{
				strcpy(parabuf, bans->banstring);
				paralen = namlen;
				*cp++ = 'b';
			}
			count = 1;
		}
	}

	/* then send exceptions and quiets */
	for(excepts=chptr->excepts;excepts;excepts=excepts->next)
	{
		if (paralen + (namlen = strlen(excepts->estring)) + 10 < 200)
		{
			*(parabuf+paralen) = ' ';
			strcpy(parabuf+paralen+1, excepts->estring);
			paralen += namlen + 1;
			count++;
			*cp++ = 'e';
		}
		else if (*parabuf)
			send = 1;
		if (count == 4)
			send = 1;
		if (send)
		{
			*cp = '\0';
			send_prefix_to_client(meptr, cptr, "MODE %s %s %s",
					chptr->name, modebuf, parabuf);
			send = 0;
			*parabuf = '\0';
			paralen = 0;
			cp = modebuf;
			*cp++ = '+';
			if (count != 4)
			{
				strcpy(parabuf, excepts->estring);
				paralen = namlen;
				*cp++ = 'e';
			}
			count = 1;
		}
	}

	for(quiets=chptr->quiets;quiets;quiets=quiets->next)
	{
		if (paralen + (namlen = strlen(quiets->qstring)) + 10 < 200)
		{
			*(parabuf+paralen) = ' ';
			strcpy(parabuf+paralen+1, quiets->qstring);
			paralen += namlen + 1;
			count++;
			*cp++ = 'q';
		}
		else if (*parabuf)
			send = 1;
		if (count == 4)
			send = 1;
		if (send)
		{
			*cp = '\0';
			send_prefix_to_client(meptr, cptr, "MODE %s %s %s",
					chptr->name, modebuf, parabuf);
			send = 0;
			*parabuf = '\0';
			paralen = 0;
			cp = modebuf;
			*cp++ = '+';
			if (count != 4)
			{
				strcpy(parabuf, quiets->qstring);
				paralen = namlen;
				*cp++ = 'q';
			}
			count = 1;
		}
	}


	if (*parabuf)
	{
		*cp = '\0';
		send_prefix_to_client(meptr, cptr, "MODE %s %s %s",
				chptr->name, modebuf, parabuf);
	}
}
#endif

static void _umode_build_change(IRCD_CLIENT *pptr, char *buf, int oldmodes, int sendmask)
{
	int		*s;
	int		flag;
	char	plusminus = '\0';
	char	*modeptr = buf;

	if ((oldmodes & sendmask) != (pptr->c_umodes & sendmask))
	{
		for(s=_user_modes;(flag=*s);s+=2)
		{
			if (MyClient(pptr) && !(flag & sendmask))
				continue;
			if ((flag & oldmodes) && !(pptr->c_umodes & flag))
			{
				if (plusminus != '-')
				{
					plusminus = '-';
					*modeptr++ = plusminus;
				}
				*modeptr++ = *(s+1);
			}
			else if (!(flag & oldmodes) && (pptr->c_umodes & flag))
			{
				if (plusminus != '+')
				{
					plusminus = '+';
					*modeptr++ = plusminus;
				}
				*modeptr++ = *(s+1);
			}
		}
	}
	*modeptr = '\0';
}

static void _umode_build(IRCD_CLIENT *pptr, char *buf, int mask)
{
	int		*s;
	int		flag;

	*buf++ = '+';
	for(s=_user_modes;(flag=*s);s+=2)
	{
		if (!(flag & mask))
			continue;
		if (pptr->c_umodes & flag)
			*buf++ = *(s+1);
	}
	*buf = '\0';
}

#ifdef HMC
static int m_umode(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *nick, char *prefix, char *args, char *hmc)
#else
static int m_umode(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *nick, char *prefix, char *args)
#endif
{
	char			*modes 	= ircd_nextarg(&args, 0);
	char			plusminus = '+';
	IRCD_CLIENT		*acptr;
	char			ubuf[64];
	int				oldmodes;
	char			*ptr;
	int				*i;
	int			showflags	= 0;

	if (!IsServer(lptr) && !IsReg(pptr))
	{
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NOTREGISTERED, NULL));
		return 3;
	}

	if (!nick || !*nick) /* should always be false, checked in m_mode() */
	{
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NEEDMOREPARAMS,
				pptr->name), "UMODE");
		return 3;
	}

	acptr = hash_find_client(nick);
	if (!acptr || IsServer(acptr))
	{
		if (IsLocal(pptr))
		{
			send_prefix_to_client(meptr, pptr,
				IRCD_REPLY(ERR_NOSUCHCHANNEL, pptr->name),
				nick);
		}
		return 3;
	}

	if (acptr != pptr)
	{
		send_prefix_to_client(meptr, pptr,
				IRCD_REPLY(ERR_USERSDONTMATCH, pptr->name));
		return 3;
	}

	if (!modes || !*modes)
	{
		_umode_build(acptr, ubuf, -1);
		send_prefix_to_client(meptr, pptr,
				IRCD_REPLY(RPL_UMODEIS, pptr->name), ubuf);
		return 3;
	}

	oldmodes = pptr->c_umodes;

	for(ptr=modes;*ptr;ptr++)
	{
		switch(*ptr)
		{
			case '+':
			case '-':
				plusminus = *ptr;
				break;
			default:
				for(i=_user_modes;*i;i+=2)
					if (*ptr == *(i+1))
						break;
				if (!*i)
				{
					send_prefix_to_client(meptr, pptr,
							IRCD_REPLY(ERR_UMODEUNKNOWNFLAG, pptr->name), *ptr);
					continue;
				}
				if (MyClient(pptr))
				{
#ifdef OPER_GOD_MODE /* made originally as a debugging feature on glines, left in because someone wanted it on their net;
		      * this is an "undocumented feature" which will probably be pulled completely later */
						/* +r */
					if(*i == CLIENT_UMODES_REGISTERED)
						continue;				/* +x */
					if(!IsAnOper(pptr) && MyClient(pptr) && (*i == CLIENT_UMODES_SERVICE))
						continue;
#else
						/* +x */			/* +r */
					if(*i == CLIENT_UMODES_SERVICE || *i == CLIENT_UMODES_REGISTERED)
						continue;
#endif
				}
				if (plusminus == '+')
					pptr->c_umodes |= *i;
				else
					pptr->c_umodes &= ~(*i);
				break;
		}
	}

	if (pptr->c_umodes == oldmodes)
		return 3;

	if (!(oldmodes & CLIENT_UMODES_OPER) &&
					(pptr->c_umodes & CLIENT_UMODES_OPER))
	{
		if (IsServer(lptr))
			Globalinfo->g_opers++;
		else
			pptr->c_umodes &= ~CLIENT_UMODES_OPER;
	}
	else if (!(pptr->c_umodes & CLIENT_UMODES_OPER) &&
				(oldmodes & CLIENT_UMODES_OPER))
	{
		/* deopering; remove all oper-only modes to return them to normal */

		if(pptr->c_umodes & CLIENT_UMODES_FMODE)
			pptr->c_umodes &= ~CLIENT_UMODES_FMODE;
                if(pptr->c_umodes & CLIENT_UMODES_KMODE)
		        pptr->c_umodes &= ~CLIENT_UMODES_KMODE;
	        if(pptr->c_umodes & CLIENT_UMODES_WMODE)
		        pptr->c_umodes &= ~CLIENT_UMODES_WMODE;
		if(pptr->c_umodes & CLIENT_UMODES_ZMODE)
		        pptr->c_umodes &= ~CLIENT_UMODES_ZMODE;
		if(pptr->c_umodes & CLIENT_UMODES_YMODE)
		        pptr->c_umodes &= ~CLIENT_UMODES_YMODE;
		if(pptr->c_umodes & CLIENT_UMODES_CMODE)
			pptr->c_umodes &= ~CLIENT_UMODES_CMODE;
		if(pptr->c_umodes & CLIENT_UMODES_NMODE)
		        pptr->c_umodes &= ~CLIENT_UMODES_NMODE;
#ifdef OPER_GOD_MODE /* see comments in m_mode() about this */
		if(pptr->c_umodes & CLIENT_UMODES_SERVICE)
		        pptr->c_umodes &= ~CLIENT_UMODES_SERVICE;
#endif
		/* strip any of the clients oper only flags */
		if(MyClient(pptr))
		{
			IRCD_USERFLAG 	*flptr;
			for(flptr = _user_flags;flptr->flagstr;flptr++)
		            if (flptr->opersonly && (pptr->c_nflags & flptr->flag))
				    pptr->c_nflags &= ~flptr->flag;
			showflags = 1;

			pptr->c_wl = pptr->c_old_wl;
			pptr->c_oline_flags = 0;
		}
			
		Globalinfo->g_opers--;
		assert(pptr->c_oslist);
		del_client_from_oslist(pptr->c_oslist, pptr);
		add_client_to_oslist(Clients, pptr);

	}

/* can't get this one from remote server...*/
	if (!(oldmodes & CLIENT_UMODES_LOCOP) &&
					(pptr->c_umodes & CLIENT_UMODES_LOCOP))
	{
		if (!IsServer(lptr))
			pptr->c_umodes &= ~CLIENT_UMODES_LOCOP;
	}

	if (!(oldmodes & CLIENT_UMODES_IMODE) &&
			(pptr->c_umodes & CLIENT_UMODES_IMODE))
	{
		if (IsLocal(pptr))
		{
			Globalinfo->m_invis++;
			pptr->c_nflags |= CLIENT_NFLAGS_INVISIBLE; /* yes, the flags are usually for notices and things, but
								    * adding flags for invisible and deaf umodes saves people asking
								    * the opers how things work --myriad */
			showflags = 1;
		}

		Globalinfo->g_invis++;
	}
	else if (!(pptr->c_umodes & CLIENT_UMODES_IMODE) &&
			(oldmodes & CLIENT_UMODES_IMODE))
	{
		if (IsLocal(pptr))
		{
			Globalinfo->m_invis--;
			pptr->c_nflags &= ~CLIENT_NFLAGS_INVISIBLE;
			showflags = 1;
		}

		Globalinfo->g_invis--;
		assert(pptr->c_oslist);
		del_client_from_oslist(pptr->c_oslist, pptr);
		add_client_to_oslist(Clients, pptr);
	}

       if (!(oldmodes & CLIENT_UMODES_DEAF) &&
                         (pptr->c_umodes & CLIENT_UMODES_DEAF))
        {
		if(IsLocal(pptr))
		{
                	pptr->c_nflags |= CLIENT_NFLAGS_DEAF;
        		showflags = 1;
		}
        }
        else if ((oldmodes & CLIENT_UMODES_DEAF) &&
                             !(pptr->c_umodes & CLIENT_UMODES_DEAF))
        {
		if(IsLocal(pptr))
		{
                	pptr->c_nflags &= ~CLIENT_NFLAGS_DEAF;
        		showflags = 1;
		}
        }

	if (!(oldmodes & CLIENT_UMODES_ONLYREGMSGS) &&
			(pptr->c_umodes & CLIENT_UMODES_ONLYREGMSGS))
	{
		if(IsLocal(pptr))
		{
			pptr->c_nflags |= CLIENT_NFLAGS_ONLYREGISTEREDMSGS;
			showflags = 1;
			if(!IsRegistered(pptr))
				send_prefix_to_client(meptr, pptr, "NOTICE %s :*** Note that umode +R only has an effect if you are using a registered nick", 
								pptr->name);
		}
	}
	else if ((oldmodes & CLIENT_UMODES_ONLYREGMSGS) &&
				!(pptr->c_umodes & CLIENT_UMODES_ONLYREGMSGS))
	{
		if(IsLocal(pptr))
		{
			pptr->c_nflags &= ~CLIENT_NFLAGS_ONLYREGISTEREDMSGS;
			showflags = 1;
		}
	}

	if (IsLocal(pptr))
	{
		/* int showflags = 0; */

		if (!(oldmodes & CLIENT_UMODES_ZMODE) &&
				(pptr->c_umodes & CLIENT_UMODES_ZMODE))
		{
			if (IsAnOper(pptr))
			{
				pptr->c_nflags |= CLIENT_NFLAGS_OWALLOPS;
				showflags = 1;
			}
			else
			{
				pptr->c_umodes &= ~CLIENT_UMODES_ZMODE;
			}
		}
		else if ((oldmodes & CLIENT_UMODES_ZMODE) &&
				!(pptr->c_umodes & CLIENT_UMODES_ZMODE))
		{
			pptr->c_nflags &= ~CLIENT_NFLAGS_OWALLOPS;
			showflags = 1;
		}
		/* umode +s is no longer restricted because it shows NOTHING beyond a "highest connection count" notice */
		if (!(oldmodes & CLIENT_UMODES_SMODE) &&
				(pptr->c_umodes & CLIENT_UMODES_SMODE))
		{
				pptr->c_nflags |= CLIENT_NFLAGS_SNOTICES;
				showflags = 1;
		}
		else if ((oldmodes & CLIENT_UMODES_SMODE) &&
				!(pptr->c_umodes & CLIENT_UMODES_SMODE))
		{
			pptr->c_nflags &= ~CLIENT_NFLAGS_SNOTICES;
			showflags = 1;
		}
		if (!(oldmodes & CLIENT_UMODES_FMODE) &&
				(pptr->c_umodes & CLIENT_UMODES_FMODE))
		{
			if(IsAnOper(pptr))
			{
				pptr->c_nflags |= CLIENT_NFLAGS_SKILLS;
				showflags = 1;
			}
			else
				pptr->c_umodes &= ~CLIENT_UMODES_FMODE;
		}
		else if ((oldmodes & CLIENT_UMODES_FMODE) &&
				!(pptr->c_umodes & CLIENT_UMODES_FMODE))
		{
			pptr->c_nflags &= ~CLIENT_NFLAGS_SKILLS;
			showflags = 1;
		}
		if (!(oldmodes & CLIENT_UMODES_KMODE) &&
				(pptr->c_umodes & CLIENT_UMODES_KMODE))
		{
			if(IsAnOper(pptr))
			{
				pptr->c_nflags |= CLIENT_NFLAGS_OKILLS;
				showflags = 1;
			}
			else
				pptr->c_umodes &= ~CLIENT_UMODES_KMODE;
		}
		else if ((oldmodes & CLIENT_UMODES_KMODE) &&
				!(pptr->c_umodes & CLIENT_UMODES_KMODE))
		{
			pptr->c_nflags &= ~CLIENT_NFLAGS_OKILLS;
			showflags = 1;
		}
		if (!(oldmodes & CLIENT_UMODES_WMODE) &&
				(pptr->c_umodes & CLIENT_UMODES_WMODE))
		{
			if(IsAnOper(pptr))
			{
				pptr->c_nflags |= CLIENT_NFLAGS_SWALLOPS;
				showflags = 1;
			}
			else
				pptr->c_umodes &= ~CLIENT_UMODES_WMODE;
		}
		else if ((oldmodes & CLIENT_UMODES_WMODE) &&
				!(pptr->c_umodes & CLIENT_UMODES_WMODE))
		{
			pptr->c_nflags &= ~CLIENT_NFLAGS_SWALLOPS;
			showflags = 1;
		}
		if (!(oldmodes & CLIENT_UMODES_CMODE) &&
				(pptr->c_umodes & CLIENT_UMODES_CMODE))
		{
			if (IsOper(pptr))
			{
				pptr->c_nflags |= CLIENT_NFLAGS_CLICONNECTS|CLIENT_NFLAGS_CLIDISCONNECTS;
				showflags = 1;
			}
			else
			{
				pptr->c_nflags &= ~(CLIENT_NFLAGS_CLICONNECTS|CLIENT_NFLAGS_CLIDISCONNECTS);
				pptr->c_umodes &= ~CLIENT_UMODES_CMODE;
			}
		}
		else if ((oldmodes & CLIENT_UMODES_CMODE) &&
				!(pptr->c_umodes & CLIENT_UMODES_CMODE))
		{
			pptr->c_nflags &= ~(CLIENT_NFLAGS_CLICONNECTS|CLIENT_NFLAGS_CLIDISCONNECTS);
			showflags = 1;
		}
		if (!(oldmodes & CLIENT_UMODES_NMODE) &&
				(pptr->c_umodes & CLIENT_UMODES_NMODE))
		{
			if (IsOper(pptr))
			{
				pptr->c_nflags |= CLIENT_NFLAGS_NICKCHANGES;
				showflags = 1;
			}
			else
			{
				pptr->c_nflags &= ~CLIENT_NFLAGS_NICKCHANGES;
				pptr->c_umodes &= ~CLIENT_UMODES_NMODE;
			}
		}
		else if ((oldmodes & CLIENT_UMODES_NMODE) &&
				!(pptr->c_umodes & CLIENT_UMODES_NMODE))
		{
			pptr->c_nflags &= ~CLIENT_NFLAGS_NICKCHANGES;
			showflags = 1;
		}
		if (!(oldmodes & CLIENT_UMODES_YMODE) &&
				(pptr->c_umodes & CLIENT_UMODES_YMODE))
		{
			if (IsOper(pptr))
			{
				pptr->c_nflags |= CLIENT_NFLAGS_STATSNOTICES;
				showflags = 1;
			}
			else
			{
				pptr->c_nflags &= ~CLIENT_NFLAGS_STATSNOTICES;
				pptr->c_umodes &= ~CLIENT_UMODES_YMODE;
			}
		}
		else if ((oldmodes & CLIENT_UMODES_YMODE) &&
				!(pptr->c_umodes & CLIENT_UMODES_YMODE))
		{
			pptr->c_nflags &= ~CLIENT_NFLAGS_STATSNOTICES;
			showflags = 1;
		}

		if (showflags)
			m_flags(lptr, pptr, prefix, NULL);

		if (oldmodes == pptr->c_umodes)
			return 3;
	}

#ifdef HMC
	if(hmc && !MyClient(pptr))
	{
		hmc++; 
		set_hmc_remote(pptr, hmc);
	}

	if(MyClient(pptr))	
		set_hmc(pptr);
#endif
	
	_umode_build_change(pptr, ubuf, oldmodes, SEND_UMODES);
	if (*ubuf)
#ifdef HMC
		send_to_servers(lptr, ":%s MODE %s %s !%s", pptr->name, pptr->name, ubuf, hmc_decide(pptr));
#else
		send_to_servers(lptr, ":%s MODE %s %s", pptr->name, pptr->name, ubuf);
#endif
	_umode_build_change(pptr, ubuf, oldmodes, -1);
#if 0
	if (MyClient(pptr))
		send_prefix_to_client(pptr, pptr, "MODE %s %s", pptr->name, buf);
/* umodes only send client's nick, not full n!u@h */
#else
	if (MyClient(pptr))
		send_to_client(pptr, ":%s MODE %s %s", pptr->name, pptr->name, ubuf);
#endif

	return 3;
}

int m_omode(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	if(MyClient(pptr))
	{
		if(!IsAnOper(pptr))
		{
			send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NOPRIVILEGES, pptr->name));
			return 3;
		}
		if(!(pptr->c_oline_flags & OLINE_FLAGS_CANOMODE))
		{
			send_prefix_to_client(meptr, pptr, "NOTICE %s :You are not allowed to use OMODE.", pptr->name);
			return 3;
		}
	}

	return do_mode(lptr, pptr, prefix, args, 1);
}
int m_mode(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	return do_mode(lptr, pptr, prefix, args, 0);
}

int do_mode(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args, int omode)
{
	IRCD_CLIENT		*tptr;
	IRCD_CHANNEL	*chptr;
	IRCD_CHANNELLIST	*chlptr;
	IRCD_CHANNELLIST	*chlptr2;
	char			*channel = ircd_nextarg(&args, 0);
	char			*modes = ircd_nextarg(&args, 0);
	char			*modeargs = ircd_nextarg(&args, NEXTARG_FLAGS_LASTARG);
	char			*ptr;
	char			plusminus = '+';
	char			newban[NICKLEN+USERLEN+HOSTLEN+6];
	char			newexcept[NICKLEN+USERLEN+HOSTLEN+6];
	char			newquiet[NICKLEN+USERLEN+HOSTLEN+6];
	IRCD_MODE		*opmodes = NULL;
	IRCD_MODE		*lastopmode = NULL;
	IRCD_MODE		*banmodes = NULL;
	IRCD_MODE		*lastbanmode = NULL;
	IRCD_MODE		*exceptmodes	= NULL;
	IRCD_MODE		*lastexceptmode = NULL;
	IRCD_MODE		*quietmodes = NULL;
	IRCD_MODE		*lastquietmode = NULL;
	IRCD_MODE		*opmodeptr;
	char			parabuf[512];
	char			modebuf[512];
	char			modeargbuf[512];
	char			*modeptr = modebuf;
	char			*modeargptr = modeargbuf;
	char			*newkey = NULL;
	int				limitchange = 0;
	int				limit = 0;
	int				tlimit;
	int				keychange = 0;
	int				opcnt = 0;
	int				*i;
	int				newmodes;
	int				oldmodes;
	char			*sptr;
	char			*target;
	int				isok;
	int				ischop;
	int				isdeop;
	int				got_ovk = 0;
	IRCD_BAN		*ban;
	IRCD_EXCEPT		*except;
	IRCD_QUIET		*quiet;

	if (!IsServer(lptr) && !IsReg(pptr))
	{
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NOTREGISTERED, NULL));
		return 3;
	}
	if (!channel || !*channel)
	{
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NEEDMOREPARAMS,
				pptr->name), "MODE");
		return 3;
	}
	chptr = hash_find_channel(channel);
	if (!chptr)
#ifdef HMC
		return m_umode(lptr, pptr, channel, prefix, modes, modeargs);
#else
		return m_umode(lptr, pptr, channel, prefix, modes);
#endif

	if (!modes || !*modes)
	{
		_build_channel_modes(chptr, pptr, modebuf, parabuf);
		send_prefix_to_client(meptr, pptr,
			IRCD_REPLY(RPL_CHANNELMODEIS, pptr->name), chptr->name,
			modebuf, parabuf);
		send_prefix_to_client(meptr, pptr,
			IRCD_REPLY(RPL_CREATIONTIME, pptr->name), chptr->name,
			chptr->creation);
		return 3;
	}

	chlptr = IsMember(chptr, pptr);

	ischop = IsServer(pptr) || (omode) || 
				(chlptr && (chlptr->clptr->modes & CHANNEL_MODES_CHANOP));
	isdeop = !ischop && !IsServer(pptr) &&
				(chlptr && (chlptr->clptr->modes & CHANNEL_MODES_DEOPPED));
	isok = ischop || (!isdeop && IsServer(lptr) && chptr->creation) || !MyClient(pptr);

	oldmodes = newmodes = chptr->modes;

	for(ptr=modes;ptr && *ptr;)
	{
		switch(*ptr)
		{
			case '+':
			case '-':
				plusminus = *ptr;
				break;
			case 'o':
			case 'v':
				target = ircd_next_subarg_space(&modeargs);
				if (!target || !*target)
				{
#if 0
					send_prefix_to_client(meptr, pptr,
						IRCD_REPLY(ERR_NEEDMOREPARAMS, pptr->name),
						"MODE-ov");
#endif
					break;
				}
				if (IsClient(lptr) && !omode)
				{
					if (!chlptr)
					{
						send_prefix_to_client(meptr, pptr,
							IRCD_REPLY(ERR_NOTONCHANNEL, pptr->name),
							chptr->name);
						break;
					}
					if (!ischop)
					{
						send_prefix_to_client(meptr, pptr,
							IRCD_REPLY(ERR_CHANOPRIVSNEEDED, pptr->name),
							chptr->name);
						break;
					}

					if (opcnt++ >= MAXMODEPARAMS)
						break;
				}
				tptr = hash_find_chasing(target);
				if (!tptr)
					break;
				if (!IsMember(chptr, tptr) && !omode)
				{
					send_prefix_to_client(meptr, pptr,
						IRCD_REPLY(ERR_USERNOTINCHANNEL, pptr->name),
						tptr->name, chptr->name);
					break;
				}
				if (IsService(tptr) && !omode && (plusminus == '-') && (*ptr == 'o'))
				{
						send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_ITSASERVICE, pptr->name),
								tptr->name, chptr->name);
						break;
				}
				if ((tptr == lptr) && (plusminus == '+') && (*ptr == 'o') && !omode && MyClient(tptr))
				{
					break;
				}
				if (!got_ovk)
					got_ovk = *ptr;
				else if (IsClient(lptr) && (got_ovk != *ptr))
					break;
				/* keep track of 'tptr' and what mode change */
				opmodeptr = _ircdmode_create(&opmodes, &lastopmode);
				opmodeptr->cptr = tptr;
				opmodeptr->mode = *ptr;
				opmodeptr->plusminus = plusminus;
				break;
			case 'k':
				target = ircd_next_subarg_space(&modeargs);
				if (!target || !*target)
				{
					if (MyClient(pptr))
						send_prefix_to_client(meptr, pptr,
							IRCD_REPLY(ERR_NEEDMOREPARAMS, pptr->name),
							"MODE-k");
					break;
				}
				if (!IsServer(pptr) && !chlptr && !omode)
				{
					send_prefix_to_client(meptr, pptr,
						IRCD_REPLY(ERR_NOTONCHANNEL, pptr->name),
						chptr->name);
					break;
				}
				if (!ischop)
				{
					send_prefix_to_client(meptr, pptr,
						IRCD_REPLY(ERR_CHANOPRIVSNEEDED, pptr->name),
						chptr->name);
					break;
				}
				if (keychange)
					break;
				else
				{
					register u_char		*s;

					if (IsClient(lptr))
					{
						register u_char		*d;
						register u_char		ch;

						for(d=s=(u_char *)target;(ch=*s);s++)
						{
							ch &= 0x7f;
							if ((ch != ':') && (ch > 0x20))
							{
								*d++ = ch;
							}
						}
						*d = '\0';
					}
					else
					{
						for(s=(u_char *)target;*s;s++)
						{
							*s &= 0x7f;
							if ((*s == ' ') || (*s == '\t') || (*s == '\n'))
							{
								*s = '\0';
								break;
							}
						}
					}
				}
				if (!*target || ((IsClient(lptr) &&
									(opcnt++ >= MAXMODEPARAMS))))
					break;
				if (!got_ovk)
					got_ovk = *ptr;
				else if (IsClient(lptr) && (*ptr != got_ovk))
					break;
				keychange = 1;
				newkey = chptr->key;
				if (plusminus == '+')
				{
					if (chptr->key && IsClient(lptr))
						send_prefix_to_client(meptr, pptr,
							IRCD_REPLY(ERR_KEYSET, pptr->name),
							chptr->name);
					else if (!chptr->key || IsServer(lptr))
						newkey = target;
				}
				else
				{
					if (IsServer(lptr) ||
							(chptr->key && !mycmp(chptr->key, target)))
					{
						newkey = NULL;
					}
				}
				break;
			case 'b':
				target = ircd_next_subarg_space(&modeargs);
				if (!target || !*target)
				{
					/* return banlist */
					if(HideCBans && chptr)
					{
						if(!IsAnOper(pptr) && !IsMember(chptr, pptr))
						{
							send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NOTONCHANNEL, pptr->name), chptr->name);
							break;
						}
						if(!IsAnOper(pptr) && !(chlptr->clptr->modes & CHANNEL_MODES_CHANOP))
						{
							send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_CHANOPRIVSNEEDED, pptr->name), chptr->name);
							break;
						}

                                        }

					for(ban=chptr->bans;ban;ban=ban->next)
					{
						send_prefix_to_client(meptr, pptr, 
							IRCD_REPLY(RPL_BANLIST, pptr->name),
							chptr->name,
							ban->banstring,
							ban->who,
							ban->bantime);
					}
					send_prefix_to_client(meptr, pptr, 
						IRCD_REPLY(RPL_ENDOFBANLIST, pptr->name),
						chptr->name);
					break;
				}
				if (!IsServer(pptr) && !chlptr && !omode)
				{
					send_prefix_to_client(meptr, pptr,
						IRCD_REPLY(ERR_NOTONCHANNEL, pptr->name),
						chptr->name);
					break;
				}
				if (!ischop)
				{
					send_prefix_to_client(meptr, pptr,
						IRCD_REPLY(ERR_CHANOPRIVSNEEDED, pptr->name),
						chptr->name);
					break;
				}
				if (IsClient(lptr) && opcnt++ >= MAXMODEPARAMS)
					break;
				opmodeptr = _ircdmode_create(&banmodes, &lastbanmode);
				opmodeptr->mode = *ptr;
				opmodeptr->plusminus = plusminus;
				opmodeptr->banstring = target;
				break;
			case 'e':
				target = ircd_next_subarg_space(&modeargs);
				if (!target || !*target)
				{
                                        if(HideCExcepts && chptr)
                                        {
                                                if(!IsAnOper(pptr) && !IsMember(chptr, pptr))
                                                {
                                                        send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NOTONCHANNEL, pptr->name), chptr->name);
                                                        break;
                                                }
                                                if(!IsAnOper(pptr) && !(chlptr->clptr->modes & CHANNEL_MODES_CHANOP))
                                                {
                                                        send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_CHANOPRIVSNEEDED, pptr->name), chptr->name);
                                                        break;
                                                }
                                                
                                        }

					for(except=chptr->excepts;except;except=except->next)
					{
						send_prefix_to_client(meptr, pptr, 
							IRCD_REPLY(RPL_EXCEPTLIST, pptr->name),
							chptr->name,
							except->estring,
							except->who,
							except->etime);
					}
					send_prefix_to_client(meptr, pptr, 
						IRCD_REPLY(RPL_ENDOFEXCEPTLIST, pptr->name),
						chptr->name);
					break;
				}
				if (!IsServer(pptr) && !chlptr && !omode)
				{
					send_prefix_to_client(meptr, pptr,
						IRCD_REPLY(ERR_NOTONCHANNEL, pptr->name),
						chptr->name);
					break;
				}
				if (!ischop)
				{
					send_prefix_to_client(meptr, pptr,
						IRCD_REPLY(ERR_CHANOPRIVSNEEDED, pptr->name),
						chptr->name);
					break;
				}
				if (IsClient(lptr) && opcnt++ >= MAXMODEPARAMS)
					break;
				opmodeptr = _ircdmode_create(&exceptmodes, &lastexceptmode);
				opmodeptr->mode = *ptr;
				opmodeptr->plusminus = plusminus;
				opmodeptr->exceptstring = target; /* could probably share this with 'banstring', might do so later */
				break;
			case 'q':
				target = ircd_next_subarg_space(&modeargs);
				if (!target || !*target)
				{
                                        if(HideCQuiets && chptr)
                                        {
                                                if(!IsAnOper(pptr) && !IsMember(chptr, pptr))
                                                {
                                                        send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NOTONCHANNEL, pptr->name), chptr->name);
                                                        break;
                                                }
                                                if(!IsAnOper(pptr) && !(chlptr->clptr->modes & CHANNEL_MODES_CHANOP))
                                                {
                                                        send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_CHANOPRIVSNEEDED, pptr->name), chptr->name);
                                                        break;
                                                }
                                                
                                        }
					for(quiet=chptr->quiets;quiet;quiet=quiet->next)
					{
						send_prefix_to_client(meptr, pptr, 
							IRCD_REPLY(RPL_QUIETLIST, pptr->name),
							chptr->name,
							quiet->qstring,
							quiet->who,
							quiet->qtime);
					}
					send_prefix_to_client(meptr, pptr, 
						IRCD_REPLY(RPL_ENDOFQUIETLIST, pptr->name),
						chptr->name);
					break;
				}
				if (!IsServer(pptr) && !chlptr && !omode)
				{
					send_prefix_to_client(meptr, pptr,
						IRCD_REPLY(ERR_NOTONCHANNEL, pptr->name),
						chptr->name);
					break;
				}
				if (!ischop)
				{
					send_prefix_to_client(meptr, pptr,
						IRCD_REPLY(ERR_CHANOPRIVSNEEDED, pptr->name),
						chptr->name);
					break;
				}
				if (IsClient(lptr) && opcnt++ >= MAXMODEPARAMS)
					break;
				opmodeptr = _ircdmode_create(&quietmodes, &lastquietmode);
				opmodeptr->mode = *ptr;
				opmodeptr->plusminus = plusminus;
				opmodeptr->quietstring = target;
				break;
			case 'l':
				tlimit = 0;
				if (plusminus == '+')
				{
					target = ircd_next_subarg_space(&modeargs);
					if (!target || !*target || !(tlimit = atoi(target)))
					{
						send_prefix_to_client(meptr, pptr,
							IRCD_REPLY(ERR_NEEDMOREPARAMS, pptr->name),
							"MODE-l");
						break;
					}
				}
				if (!IsServer(pptr) && !chlptr && !omode)
				{
					send_prefix_to_client(meptr, pptr,
						IRCD_REPLY(ERR_NOTONCHANNEL, pptr->name),
						chptr->name);
					break;
				}
				if (!ischop)
				{
					send_prefix_to_client(meptr, pptr,
						IRCD_REPLY(ERR_CHANOPRIVSNEEDED, pptr->name),
						chptr->name);
					break;
				}
				if (limitchange++)
					break;
				limit = tlimit;
				break;
			case 'O':
				if(!IsAnOper(pptr))
				{
					send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NOPRIVILEGES, pptr->name));
					break;
				}
			case 'i':
			case 's':
			case 'p':
			case 't':
			case 'n':
			case 'm':
			case 'R':
				if (!IsServer(pptr) && !chlptr && !omode)
				{
					send_prefix_to_client(meptr, pptr,
						IRCD_REPLY(ERR_NOTONCHANNEL, pptr->name),
						chptr->name);
					break;
				}
				if (!ischop)
				{
					send_prefix_to_client(meptr, pptr,
						IRCD_REPLY(ERR_CHANOPRIVSNEEDED, pptr->name),
						chptr->name);
					break;
				}
				for(i=_channel_modes;*i;i+=2)
				{
					if (*(i+1) != *ptr)
						continue;
					if (plusminus == '+')
						newmodes |= *i;
					else
						newmodes &= ~(*i);
				}
				break;
			default:
				if (IsClient(pptr))
					send_prefix_to_client(meptr, pptr,
						IRCD_REPLY(ERR_UNKNOWNMODE, pptr->name), *ptr);
				break;
		}
		if (!*++ptr)
			ptr = ircd_next_subarg_space(&modeargs);
	}

    /* if it's a non-deopped non-op on a TS channel from a TS
    ** server, allow only ops and deops and ignore everything
    ** else (to avoid having desynched ops) -orabidoo
    */
	if (!ischop && isok)
	{
		newmodes = oldmodes;
		limitchange = 0;
		keychange = 0;
		_ircdmode_free(&banmodes);
		_ircdmode_free(&exceptmodes);
		_ircdmode_free(&quietmodes);
	}

	plusminus = '\0';

	if (newmodes != oldmodes)
	{
		for(i=_channel_modes;*i;i+=2)
		{
			if ((*i & newmodes) && !(*i & oldmodes))
			{
				if (plusminus != '+')
				{
					plusminus = '+';
					*modeptr++ = plusminus;
				}
				*modeptr++ = *(i+1);
				chptr->modes |= *i;
			}
		}

		for(i=_channel_modes;*i;i+=2)
		{
			if ((*i & oldmodes) && !(*i & newmodes))
			{
				if (plusminus != '-')
				{
					plusminus = '-';
					*modeptr++ = plusminus;
				}
				*modeptr++ = *(i+1);
				chptr->modes &= ~(*i);
			}
		}
	}

	if (limitchange)
	{
		if (limit && (chptr->limit != limit))
		{
			char	buf[20];

			if (plusminus != '+')
			{
				plusminus = '+';
				*modeptr++ = plusminus;
			}
			*modeptr++ = 'l';
			sprintf(buf, "%-15d", limit);
			if (modeargptr != modeargbuf)
				*modeargptr++ = ' ';
			for(sptr=buf;*sptr && *sptr != ' ';)
					*modeargptr++ = *sptr++;
			chptr->limit = limit;
		}
		if (!limit && chptr->limit)
		{
			if (plusminus != '-')
			{
				plusminus = '-';
				*modeptr++ = plusminus;
			}
			*modeptr++ = 'l';
			chptr->limit = 0;
		}
	}

	/* build +ov */

	for(opmodeptr=opmodes;opmodeptr;opmodeptr=opmodeptr->next)
	{
		if ((modeargptr-modeargbuf) +
				strlen(opmodeptr->cptr->name) + 2 >= sizeof(modeargbuf))
			break;
		if (plusminus != opmodeptr->plusminus)
		{
			plusminus = opmodeptr->plusminus;
			*modeptr++ = plusminus;
		}
		chlptr2 = IsMember(chptr, opmodeptr->cptr);
		if (opmodeptr->mode == 'o')
		{
			if (IsServer(pptr))
			{
				send_to_opers("Server %s doing mode %co and removing TS on %s",
					pptr->name, plusminus, chptr->name);
				chptr->creation = 0;
			}

			if (isdeop && (plusminus == '+'))
			{
				if (chlptr2 && !(chlptr2->clptr->modes & CHANNEL_MODES_CHANOP))
					chlptr2->clptr->modes |= CHANNEL_MODES_DEOPPED;
			}
		}
		if (isok)
		{
			isok = _channel_change_opflag(chptr, chlptr2->clptr,
							opmodeptr, IsClient(lptr) /* , pptr->c_oline_forcing */ );
			if (isok || IsServer(lptr))
			{
				*modeptr++ = opmodeptr->mode;
				if (modeargptr != modeargbuf)
					*modeargptr++ = ' ';
				for(sptr=opmodeptr->cptr->name;*sptr;)
					*modeargptr++ = *sptr++;
			}
		}
	}
	_ircdmode_free(&opmodes);

	/* +k */

	if (keychange)
	{
		if (chptr->key && !newkey)
		{
			if (!((modeargptr-modeargbuf) +
				strlen(chptr->key) + 2 >= sizeof(modeargbuf)))
			{
				if (plusminus != '-')
				{
					plusminus = '-';
					*modeptr++ = plusminus;
				}
				*modeptr++ = 'k';
				if (modeargptr != modeargbuf)
					*modeargptr++ = ' ';
				for(sptr=chptr->key;*sptr;)
					*modeargptr++ = *sptr++;
				MyFree(chptr->key);
				chptr->key = NULL;
			}
		}
		else if (newkey && newkey != chptr->key)
		{
			if (!((modeargptr-modeargbuf) +
				strlen(newkey) + 2 >= sizeof(modeargbuf)))
			{
				if (plusminus != '+')
				{
					plusminus = '+';
					*modeptr++ = plusminus;
				}
				*modeptr++ = 'k';
				if (modeargptr != modeargbuf)
					*modeargptr++ = ' ';
				for(sptr=newkey;*sptr;)
					*modeargptr++ = *sptr++;
				if (chptr->key)
					MyFree(chptr->key);
				chptr->key = MyStrdup(newkey);
			}
		}
	}

	/* bans */
	for(opmodeptr=banmodes;opmodeptr;opmodeptr=opmodeptr->next)
	{
		char			*user, *host;
		int				ret;
		ircd_uint32_t	net;
		ircd_uint32_t	mask;
		char			*hostptr;

		if ((user = strchr(opmodeptr->banstring, '!')) != NULL)
			*user++ = '\0';
		if ((host = strrchr(user?user:opmodeptr->banstring, '@')) != NULL)
			*host++ = '\0';
		sptr = _make_nick_user_host(opmodeptr->banstring,
								user, host, newban, &hostptr);
		ret = cidr_ip2netmask(hostptr, &net, &mask);
		if (!ret)
			net = -1;

		if ((modeargptr-modeargbuf) +
				strlen(sptr) + 2 >= sizeof(modeargbuf))
			break;

		if (((opmodeptr->plusminus == '+') &&
					!_channel_ban_add(chptr, pptr, sptr, hostptr, net, mask)) ||
				((opmodeptr->plusminus == '-') &&
					!_channel_ban_del(chptr, sptr)))
			continue;

		if (plusminus != opmodeptr->plusminus)
		{
			plusminus = opmodeptr->plusminus;
			*modeptr++ = plusminus;
		}
		*modeptr++ = opmodeptr->mode;
		if (modeargptr != modeargbuf)
			*modeargptr++ = ' ';
		for(;*sptr;)
			*modeargptr++ = *sptr++;
	}
	_ircdmode_free(&banmodes);

	/* excepts */
	for(opmodeptr=exceptmodes;opmodeptr;opmodeptr=opmodeptr->next)
	{
		char			*user, *host;
		int				ret;
		ircd_uint32_t	net;
		ircd_uint32_t	mask;
		char			*hostptr;

		if ((user = strchr(opmodeptr->exceptstring, '!')) != NULL)
			*user++ = '\0';
		if ((host = strrchr(user?user:opmodeptr->exceptstring, '@')) != NULL)
			*host++ = '\0';
		sptr = _make_nick_user_host(opmodeptr->exceptstring,
								user, host, newexcept, &hostptr);
		ret = cidr_ip2netmask(hostptr, &net, &mask);
		if (!ret)
			net = -1;

		if ((modeargptr-modeargbuf) +
				strlen(sptr) + 2 >= sizeof(modeargbuf))
			break;

		if (((opmodeptr->plusminus == '+') &&
					!_channel_except_add(chptr, pptr, sptr, hostptr, net, mask)) ||
				((opmodeptr->plusminus == '-') &&
					!_channel_except_del(chptr, sptr)))
			continue;

		if (plusminus != opmodeptr->plusminus)
		{
			plusminus = opmodeptr->plusminus;
			*modeptr++ = plusminus;
		}
		*modeptr++ = opmodeptr->mode;
		if (modeargptr != modeargbuf)
			*modeargptr++ = ' ';
		for(;*sptr;)
			*modeargptr++ = *sptr++;
	}
	_ircdmode_free(&exceptmodes);

	/* quiets */
	for(opmodeptr=quietmodes;opmodeptr;opmodeptr=opmodeptr->next)
	{
		char			*user, *host;
		int				ret;
		ircd_uint32_t	net;
		ircd_uint32_t	mask;
		char			*hostptr;

		if ((user = strchr(opmodeptr->quietstring, '!')) != NULL)
			*user++ = '\0';
		if ((host = strrchr(user?user:opmodeptr->quietstring, '@')) != NULL)
			*host++ = '\0';
		sptr = _make_nick_user_host(opmodeptr->quietstring,
								user, host, newquiet, &hostptr);
		ret = cidr_ip2netmask(hostptr, &net, &mask);
		if (!ret)
			net = -1;

		if ((modeargptr-modeargbuf) +
				strlen(sptr) + 2 >= sizeof(modeargbuf))
			break;

		if (((opmodeptr->plusminus == '+') &&
					!_channel_quiet_add(chptr, pptr, sptr, hostptr, net, mask)) ||
				((opmodeptr->plusminus == '-') &&
					!_channel_quiet_del(chptr, sptr)))
			continue;

		if (plusminus != opmodeptr->plusminus)
		{
			plusminus = opmodeptr->plusminus;
			*modeptr++ = plusminus;
		}
		*modeptr++ = opmodeptr->mode;
		if (modeargptr != modeargbuf)
			*modeargptr++ = ' ';
		for(;*sptr;)
			*modeargptr++ = *sptr++;
	}
	_ircdmode_free(&quietmodes);

	/* done with +beq .. =O */

	*modeptr = '\0';
	*modeargptr = '\0';

	if (*modebuf && isok)
	{
		if (IsClient(pptr))
			send_to_channel_butone(NULL,
				CHANNEL_SEND_CLIENTS|CHANNEL_SEND_LOCALONLY,
				chptr, pptr, "MODE %s %s %s", chptr->name, modebuf, modeargbuf);
		else
			send_to_channel_butone2(NULL,
				CHANNEL_SEND_CLIENTS|CHANNEL_SEND_LOCALONLY,
				chptr, prefix, "MODE %s %s %s", chptr->name, modebuf, modeargbuf);

		if (*chptr->name != '&')
		{
			if(!omode)
				send_to_servers(IsServer(lptr) ? lptr : NULL, ":%s MODE %s %s %s",
					prefix, chptr->name, modebuf, modeargbuf);
			else
			{
				if(MyClient(pptr))
				{
					send_wallops_to_flag(CLIENT_NFLAGS_SWALLOPS, NULL, meptr, 
						"%s[@%s] is using OMODE in %s", 
							pptr->name, meptr->name, chptr->name);
					send_to_servers(NULL, ":%s WALLOPS :%s[@%s] is using OMODE in %s",
								meptr->name, pptr->name, meptr->name, chptr->name);
				}
				send_to_servers(lptr, ":%s MODE %s %s %s",
							prefix, chptr->name, modebuf, modeargbuf);
			}
		} /* if() '&' */
	}

	return 3;
}
